from sage.all import *

def reduce_poly(poly):
    reduced_poly = poly/gcd(poly.coefficients())
    polies = []
    for fact,ex in reduced_poly.factor():
        for gen in fact.parent().gens():
            if fact.coefficient(gen)!=0:
                break
        polies.append(fact*(-1)**(fact.coefficient(gen)<0))
    return polies


def remove_y(poly,a,b,Y1,Y2,X1,X2):
    if poly.degree(Y1)==poly.degree(Y2)==0:
        return poly
    f0 = poly.constant_coefficient()
    f12 = poly.coefficient({Y1:1,Y2:1})
    f1 = poly.coefficient({Y1:1,Y2:0})
    f2 = poly.coefficient({Y2:1,Y1:0})
    curve1, curve2 = X1**3+a*X1+b,X2**3+a*X2+b
    return curve2*(curve1*2*f1*f12-2*f0*f2)**2-(f0**2+f2**2*curve2-curve1*(f1**2+f12**2*curve2))**2

def test_remove_y():
    f = 1+2*Y1+3*Y2+4*Y1*Y2
    a,b = 0,1
    expected_fprime = -256*X1**6*X2**6 - 384*X1**6*X2**3 - 224*X1**3*X2**6 - 144*X1**6 - 280*X1**3*X2**3 - 49*X2**6 - 80*X1**3 - 40*X2**3
    assert expected_fprime==remove_y(f,a,b,Y1,Y2,X1,X2)
    assert X1 == remove_y(X1,a,b,Y1,Y2,X1,X2)


def simplify_formulae(formulae):
    """transform the text into expressions, subtitute Z1=Z2=1,
    remove duplicities, common factors, remove Y1,Y2 """
    R = PolynomialRing(ZZ,['X1','X2','Z1','Z2','Y1','Y2','a','b'])
    X1,X2,Z1,Z2,Y1,Y2,a,b = R.gens()
    affine_expressions = set()
    for ex in formulae:
        try:
            polies = reduce_poly(R(ex)(Z1=1,Z2=1))
            if polies!=[]:
                affine_expressions.update(polies)
        except:
            continue
    return affine_expressions

def relation_polynomial(l,RX,field,coefs=None):
    X1,X2,X3,a,b = RX.gens()
    semaev3 = RX((X1-X2)**2*X3**2-2*((X1+X2)*(X1*X2+a)+2*b)*X3+(X1*X2-a)**2-4*b*(X1+X2))
    if l==1:
        return semaev3
    FF = PolynomialRing(field,('a','b')).fraction_field()
    a,b = FF.gens()
    E = EllipticCurve(FF,[a,b])
    ml_num = RX(E._multiple_x_numerator(l,X2))
    ml_den = RX(E._multiple_x_denominator(l,X2))
    return (X1-X3)**2*ml_num**2-2*((X1+X3)*(X1*X3+a)+2*b)*ml_num*ml_den+((X1*X3-a)**2-4*b*(X1+X3))*ml_den**2

def test_relation_polynomial(l=2):
    field = GF(137)
    RX = PolynomialRing(field,['X1','X2','X3','a','b'])
    fr = relation_polynomial(l,RX,field)
    E = EllipticCurve(field,[2,3])
    P = E.random_point()
    Q = E.random_point()
    R = -l*Q-P
    assert fr(X1=P[0],X2=Q[0],X3=R[0],a=2,b=3)==0

def relation_ideal(poly,l,field, verbose = True):
    """ Find curves for which f(P,Q)=0 means P+lQ+R=0 for some fixed R and small fixed l"""
    """ Take 3rd Semaev polynomial and reduce it modulo given polynomial from Z[x1,x2]. Find x3,a,b so that remainder is zero """
    RX = PolynomialRing(field,['X1','X2','X3','a','b'])
    X1,X2,X3,a,b = RX.gens()
    relation_poly = relation_polynomial(l,RX,field)
    RXQ = RX.quotient(poly)
    X1b,X2b,X3b,ab,bb = RXQ.gens()
    f = RXQ(relation_poly)
    f_decomposition = {} # the dictionary is of the form (deg(X1),deg(X2)):list of [coefficient,deg(X3),deg(a),deg(b)]
    for monomial in f.monomials():
        monlift = monomial.lift()
        coef = f.lift().coefficient(monlift)
        key = (monlift.degrees()[0],monlift.degrees()[1])
        value = [coef,{"degX3":monlift.degrees()[2],"dega":monlift.degrees()[3],"degb":monlift.degrees()[4]}]
        if key in f_decomposition:
            f_decomposition[key].append(value)
        else:
            f_decomposition[key]=[value]
    EquationRing=PolynomialRing(field,['X3','a','b'])
    X3,a,b = EquationRing.gens()
    equations = set()
    for key,value in f_decomposition.items():
        g = 0
        for mon in value:
            coef = ZZ(mon[0].constant_coefficient())
            degrees = mon[1]
            g+=coef*X3**degrees["degX3"]*a**degrees["dega"]*b**degrees["degb"]
        equations.add(EquationRing(g))
    equations = list(equations)
    I = EquationRing.ideal(equations)
    #if I.dimension()==-1:
    #    print("polynomial:",poly)
    #    print("dim:", I.dimension(),", ",I)
    #    print("NO SOLUTION")
    #    print('----')
    #    return
    if I.dimension()==0:
        if verbose:
            print("polynomial:", poly)
            print("dim:", I.dimension(),", ",I)
            print("SOLUTION",I.variety())
            print('----')
        return I.variety()
    if I.dimension()>0:
        if verbose:
        	print("polynomial:",poly)
        	print("dim:", I.dimension(),", ",I)
        	print("MANY SOLUTIONS")
        	print('----')
        return [None]
    return []


def load_expressions(path):
    with open(path,"r") as f:
        text = f.readlines()
    expressions = []
    for line in text:
        if "=" in line and "ZZ" not in line.split("=")[1]: # ignoring xyzz formulae for now
            expressions.append(line.split("=")[1].strip())
    return expressions
